/**
 * $Revision$
 * $Date$
 *
 * Copyright (C) 2006-2007 Jive Software. All rights reserved.
 *
 * This software is published under the terms of the GNU Public License (GPL),
 * a copy of which is included in this distribution.
 */

package org.jivesoftware.openfire.gateway.protocols.xmpp;

import org.jivesoftware.openfire.gateway.type.PresenceType;
import org.jivesoftware.openfire.gateway.type.ChatStateType;
import org.jivesoftware.openfire.gateway.type.TransportType;
import org.jivesoftware.openfire.gateway.type.TransportLoginStatus;
import org.jivesoftware.openfire.gateway.registration.Registration;
import org.jivesoftware.openfire.gateway.session.TransportSession;
import org.jivesoftware.openfire.gateway.roster.TransportBuddy;
import org.jivesoftware.openfire.gateway.protocols.xmpp.packet.*;
import org.jivesoftware.openfire.user.UserNotFoundException;
import org.jivesoftware.util.JiveGlobals;
import org.jivesoftware.util.LocaleUtils;
import org.jivesoftware.smack.*;
import org.jivesoftware.smack.provider.ProviderManager;
import org.jivesoftware.smack.filter.PacketTypeFilter;
import org.jivesoftware.smack.filter.OrFilter;
import org.jivesoftware.smack.filter.PacketExtensionFilter;
import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.util.StringUtils;
import org.xmpp.packet.JID;
import org.apache.log4j.Logger;

import java.util.Collection;
import java.util.Timer;
import java.util.TimerTask;
import java.util.ArrayList;

/**
 * Handles XMPP transport session.
 *
 * @author Daniel Henninger
 * @author Mehmet Ecevit
 */
public class XMPPSession extends TransportSession {

    static Logger Log = Logger.getLogger(XMPPSession.class);
    
    /**
     * Create an XMPP Session instance.
     *
     * @param registration Registration informationed used for logging in.
     * @param jid JID associated with this session.
     * @param transport Transport instance associated with this session.
     * @param priority Priority of this session.
     */
    public XMPPSession(Registration registration, JID jid, XMPPTransport transport, Integer priority) {
        super(registration, jid, transport, priority);
        Log.debug("Creating "+getTransport().getType()+" session for " + registration.getUsername());
        String connecthost;
        Integer connectport;
        String domain;

        connecthost = JiveGlobals.getProperty("plugin.gateway."+getTransport().getType()+".connecthost", (getTransport().getType().equals(TransportType.gtalk) ? "talk.google.com" : "jabber.org"));
        connectport = JiveGlobals.getIntProperty("plugin.gateway."+getTransport().getType()+".connectport", 5222);

        if (getTransport().getType().equals(TransportType.gtalk)) {
            domain = "gmail.com";
        }
        else {
            domain = connecthost;
        }
        
        // For differenet domains other than 'gmail.com', which is given with Google Application services
        if (registration.getUsername().indexOf("@")>-1) {
            domain = registration.getUsername().substring( registration.getUsername().indexOf("@")+1 );
        }
        
        config = new ConnectionConfiguration(connecthost, connectport, domain);
        config.setCompressionEnabled(JiveGlobals.getBooleanProperty("plugin.gateway."+getTransport().getType()+".usecompression", false));

        if (getTransport().getType().equals(TransportType.gtalk) && JiveGlobals.getBooleanProperty("plugin.gateway.gtalk.mailnotifications", true)) {
            ProviderManager.getInstance().addIQProvider(GoogleMailBoxPacket.MAILBOX_ELEMENT, GoogleMailBoxPacket.MAILBOX_NAMESPACE, new GoogleMailBoxPacket.Provider());
            ProviderManager.getInstance().addExtensionProvider(GoogleNewMailExtension.ELEMENT_NAME, GoogleNewMailExtension.NAMESPACE, new GoogleNewMailExtension.Provider());
        }
    }
    
    /*
     * XMPP connection
     */
    public XMPPConnection conn = null;
    
    /**
     * XMPP listener
     */
    private XMPPListener listener = null;

    /*
     * XMPP connection configuration
     */
    private ConnectionConfiguration config = null;

    /**
     * Timer to check for online status.
     */
    public Timer timer = new Timer();

    /**
     * Interval at which status is checked.
     */
    private int timerInterval = 60000; // 1 minute

    /**
     * Mail checker
     */
    MailCheck mailCheck;

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#logIn(org.jivesoftware.openfire.gateway.type.PresenceType, String)
     */
    public void logIn(PresenceType presenceType, String verboseStatus) {
        setPendingPresenceAndStatus(presenceType, verboseStatus);
        if (!this.isLoggedIn()) {
            listener = new XMPPListener(this);
            new Thread() {
                public void run() {
                    String userName = registration.getUsername();
                    if (userName.indexOf("@")>-1) {
                        userName = userName.substring(0,userName.indexOf("@"));
                    }

                    conn = new XMPPConnection(config);
                    try {
                        conn.connect();
                        conn.addConnectionListener(listener);
                        try {
                            conn.login(userName, registration.getPassword(), StringUtils.randomString(10), true);
                            conn.getRoster().addRosterListener(listener);
                            conn.getChatManager().addChatListener(listener);
                            // Use this to filter out anything we don't care about
                            conn.addPacketListener(listener, new OrFilter(
                                    new PacketTypeFilter(GoogleMailBoxPacket.class),
                                    new PacketExtensionFilter(GoogleNewMailExtension.ELEMENT_NAME, GoogleNewMailExtension.NAMESPACE)
                            ));
                            // Use this if we want to see -all- of the incoming packets.
//                            conn.addPacketListener(listener, new PacketFilter() {
//                                public boolean accept(Packet packet) {
//                                    return true;
//                                }
//                            });
                            setLoginStatus(TransportLoginStatus.LOGGED_IN);
                            syncUsers();

                            if (getTransport().getType().equals(TransportType.gtalk) && JiveGlobals.getBooleanProperty("plugin.gateway.gtalk.mailnotifications", true)) {
                                conn.sendPacket(new IQWithPacketExtension(getRegistration().getUsername(), new GoogleUserSettingExtension(false, true, false), IQ.Type.SET));
                                conn.sendPacket(new IQWithPacketExtension(getRegistration().getUsername(), new GoogleRelayExtension()));
                                conn.sendPacket(new IQWithPacketExtension(getRegistration().getUsername(), new GoogleMailNotifyExtension()));
                                conn.sendPacket(new IQWithPacketExtension(getRegistration().getUsername(), new GoogleSharedStatusExtension()));
                                mailCheck = new MailCheck();
                                timer.schedule(mailCheck, timerInterval, timerInterval);
                            }
                        }
                        catch (XMPPException e) {
                            Log.debug(getTransport().getType()+" user's login/password does not appear to be correct: "+getRegistration().getUsername(), e);
                            sessionDisconnectedNoReconnect(LocaleUtils.getLocalizedString("gateway.xmpp.passwordincorrect", "gateway"));
                        }
                    }
                    catch (XMPPException e) {
                        Log.debug(getTransport().getType()+" user is not able to connect: "+getRegistration().getUsername(), e);
                        sessionDisconnected(LocaleUtils.getLocalizedString("gateway.xmpp.connectionfailed", "gateway"));
                    }
                }
            }.start();
        }
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#logOut()
     */
    public void logOut() {
        cleanUp();
        sessionDisconnectedNoReconnect(null);
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#cleanUp()
     */
    public void cleanUp() {
        if (timer != null) {
            try {
                timer.cancel();
            }
            catch (Exception e) {
                // Ignore
            }
            timer = null;
        }
        if (mailCheck != null) {
            try {
                mailCheck.cancel();
            }
            catch (Exception e) {
                // Ignore
            }
            mailCheck = null;
        }
        if (conn != null) {
            try {
                conn.removeConnectionListener(listener);
            }
            catch (Exception e) {
                // Ignore
            }
            try {
                conn.getRoster().removeRosterListener(listener);
            }
            catch (Exception e) {
                // Ignore
            }
            try {
                conn.getChatManager().removeChatListener(listener);
            }
            catch (Exception e) {
                // Ignore
            }
            try {
                conn.disconnect();
            }
            catch (Exception e) {
                // Ignore
            }
        }
        conn = null;
        listener = null;
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#updateStatus(org.jivesoftware.openfire.gateway.type.PresenceType, String)
     */
    public void updateStatus(PresenceType presenceType, String verboseStatus) {
        org.jivesoftware.smack.packet.Presence p = new org.jivesoftware.smack.packet.Presence(org.jivesoftware.smack.packet.Presence.Type.available);

        if (presenceType == PresenceType.away ) {
            p.setMode(org.jivesoftware.smack.packet.Presence.Mode.away);
        }
        else if (presenceType == PresenceType.dnd ) {
            p.setMode(org.jivesoftware.smack.packet.Presence.Mode.dnd);
        }
        else if (presenceType == PresenceType.xa ) {
            p.setMode(org.jivesoftware.smack.packet.Presence.Mode.xa);
        }
        else if (presenceType == PresenceType.chat ) {
            p.setMode(org.jivesoftware.smack.packet.Presence.Mode.chat);
        }
        
        if (verboseStatus!=null) {
            p.setStatus(verboseStatus);
        }
        try {
            conn.sendPacket(p);
            setPresenceAndStatus(presenceType, verboseStatus);
        }
        catch (IllegalStateException e) {
            Log.debug("XMPP: Not connected while trying to change status.");
        }
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#addContact(org.xmpp.packet.JID, String, java.util.ArrayList)
     */
    public void addContact(JID jid, String nickname, ArrayList<String> groups) {
        String mail = getTransport().convertJIDToID(jid);
        try {
            conn.getRoster().createEntry(mail, nickname, groups.toArray(new String[0]));
            RosterEntry entry = conn.getRoster().getEntry(mail);

            getBuddyManager().storeBuddy(new XMPPBuddy(getBuddyManager(), mail, nickname, entry.getGroups(), entry));
        }
        catch (XMPPException ex) {
            Log.debug("XMPP: unable to add:"+ mail);
        }
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#removeContact(org.jivesoftware.openfire.gateway.roster.TransportBuddy)
     */
    public void removeContact(TransportBuddy contact) {
        RosterEntry user2remove;
        String mail = getTransport().convertJIDToID(contact.getJID());
        user2remove =  conn.getRoster().getEntry(mail);
        try {
            conn.getRoster().removeEntry(user2remove);
        }
        catch (XMPPException ex) {
            Log.debug("XMPP: unable to remove:"+ mail);
        }
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#updateContact(org.jivesoftware.openfire.gateway.roster.TransportBuddy)
     */
    public void updateContact(TransportBuddy contact) {
        RosterEntry user2Update;
        String mail = getTransport().convertJIDToID(contact.getJID());
        user2Update =  conn.getRoster().getEntry(mail);
        user2Update.setName(contact.getNickname());
        Collection<String> newgroups = contact.getGroups();
        for (RosterGroup group : conn.getRoster().getGroups()) {
            if (newgroups.contains(group.getName())) {
                if (!group.contains(user2Update)) {
                    try {
                        group.addEntry(user2Update);
                        newgroups.remove(group.getName());
                    }
                    catch (XMPPException e) {
                        Log.debug("XMPP: Unable to add roster item to group.");
                    }
                }
            }
            else {
                if (group.contains(user2Update)) {
                    try {
                        group.removeEntry(user2Update);
                    }
                    catch (XMPPException e) {
                        Log.debug("XMPP: Unable to delete roster item from group.");
                    }
                }
            }
        }
        for (String group : newgroups) {
            RosterGroup newgroup = conn.getRoster().createGroup(group);
            try {
                newgroup.addEntry(user2Update);
            }
            catch (XMPPException e) {
                Log.debug("XMPP: Unable to add roster item to new group.");
            }
        }
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#sendMessage(org.xmpp.packet.JID, String)
     */
    public void sendMessage(JID jid, String message) {
        Chat chat = conn.getChatManager().createChat(getTransport().convertJIDToID(jid), listener);
        try {
            chat.sendMessage(message);
        }
        catch (XMPPException e) {
            // Ignore
        }
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#sendChatState(org.xmpp.packet.JID, org.jivesoftware.openfire.gateway.type.ChatStateType)
     */
    public void sendChatState(JID jid, ChatStateType chatState) {
    }

    /**
     * @see org.jivesoftware.openfire.gateway.session.TransportSession#sendBuzzNotification(org.xmpp.packet.JID, String)
     */
    public void sendBuzzNotification(JID jid, String message) {
        Chat chat = conn.getChatManager().createChat(getTransport().convertJIDToID(jid), listener);
        try {
            Message m = new Message();
            m.setTo(getTransport().convertJIDToID(jid));
            m.addExtension(new BuzzExtension());
            chat.sendMessage(m);
        }
        catch (XMPPException e) {
            // Ignore
        }
    }
    
    public void syncUsers() {
        for (RosterEntry entry : conn.getRoster().getEntries()) {
            getBuddyManager().storeBuddy(new XMPPBuddy(getBuddyManager(), entry.getUser(), entry.getName(), entry.getGroups(), entry));
//            org.jivesoftware.smack.packet.Presence probe = new org.jivesoftware.smack.packet.Presence;
//            try {
//                conn.sendPacket(probe);
//            }
//            catch (IllegalStateException e) {
//                Log.debug("XMPP: Not connected while trying to send probe.");
//            }
        }

        try {
            getTransport().syncLegacyRoster(getJID(), getBuddyManager().getBuddies());
        }
        catch (UserNotFoundException ex) {
            ex.printStackTrace();
        }

        getBuddyManager().activate();
    }

    private class MailCheck extends TimerTask {
        /**
         * Check GMail for new mail.
         */
        public void run() {
            if (getTransport().getType().equals(TransportType.gtalk) && JiveGlobals.getBooleanProperty("plugin.gateway.gtalk.mailnotifications", true)) {
                GoogleMailNotifyExtension gmne = new GoogleMailNotifyExtension();
                gmne.setNewerThanTime(listener.getLastGMailThreadDate());
                gmne.setNewerThanTid(listener.getLastGMailThreadId());
                conn.sendPacket(new IQWithPacketExtension(getRegistration().getUsername(), gmne));
            }
        }
    }
    
}
