/**
 * $Revision$
 * $Date$
 *
 * Copyright (C) 2006-2007 Jive Software. All rights reserved.
 *
 * This software is published under the terms of the GNU Public License (GPL),
 * a copy of which is included in this distribution.
 */

package org.jivesoftware.openfire.gateway.roster;

import org.jivesoftware.openfire.roster.RosterItem;
import org.jivesoftware.openfire.gateway.type.PresenceType;
import org.xmpp.packet.JID;
import org.xmpp.packet.Presence;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.List;
import java.util.Collection;
import java.lang.ref.WeakReference;

/**
 * Transport Buddy.
 *
 * This class is intended to be extended and includes all necessary pieces for
 * syncing with a roster.  It also handles keeping current statuses and such for
 * easy retrieval and only sends presence changes upon status changes.  So that
 * the base transport can manage this list too sometimes, it is very important
 * that the specific transport implementation take into account that there may
 * buddy instances that do not know anything about any 'custom' fields the
 * transport may be implementing.
 *
 * @author Daniel Henninger
 */
public class TransportBuddy {

    static Logger Log = Logger.getLogger(TransportBuddy.class);

    /**
     * Default constructor, nothing set up.
     */
    public TransportBuddy() {
        // Nothing
    }

    /**
     * Creates a TransportBuddy instance.
     *
     * @param manager Transport buddy manager we are associated with.
     * @param contactname The legacy contact name.
     * @param nickname The legacy nickname (can be null).
     * @param groups The list of groups the legacy contact is in (can be null).
     */
    public TransportBuddy(TransportBuddyManager manager, String contactname, String nickname, Collection<String> groups) {
        this.managerRef = new WeakReference<TransportBuddyManager>(manager);
        this.jid = manager.getSession().getTransport().convertIDToJID(contactname);
        this.contactname = manager.getSession().getTransport().convertJIDToID(this.jid);
        if (nickname != null) {
            this.nickname = nickname;
        }
        else {
            this.nickname = this.contactname;
        }
        if (groups != null && !groups.isEmpty()) {
            this.groups = groups;
        }
    }

    /**
     * The transport buddy manager we are attached to.
     */
    private WeakReference<TransportBuddyManager> managerRef = null;

    public TransportBuddyManager getManager() {
        return managerRef.get();
    }

    /**
     * ID, Screenname, name, whatever the contact name is on the legacy system
     */
    public String contactname = null;

    /**
     * Converted JID for the contact, for caching purposes
     */
    public JID jid = null;

    /**
     * A nickname associated with this contact, if it exists.
     */
    public String nickname = null;

    /**
     * A group associated with this contact, if it exists.
     */
    public Collection<String> groups = new ArrayList<String>();

    /**
     * Specific requested subscription status, if desired.
     */
    public RosterItem.SubType subtype = null;

    /**
     * Specific requested ask status, if desired.
     */
    public RosterItem.AskType asktype = null;

    /**
     * Current presence status.
     */
    public PresenceType presence = PresenceType.unavailable;

    /**
     * Current verbose status.
     */
    public String verboseStatus = "";

    /**
     * Retrieves the name of the contact.
     *
     * @return Name of contact.
     */
    public String getName() {
       return contactname;
    }

    /**
     * Retrieves the JID of the contact.
     *
     *
     * @return JID of contact.
     */
    public JID getJID() {
        return jid;
    }

    /**
     * Sets the name of the contact.
     *
     * @param contactname Username of the contact.
     */
    public void setName(String contactname) {
        this.jid = getManager().getSession().getTransport().convertIDToJID(contactname);
        this.contactname = getManager().getSession().getTransport().convertJIDToID(this.jid);
    }

    /**
     * Retrieves the nickname of the contact.
     *
     * @return Nickname of contact.
     */
    public String getNickname() {
       return nickname;
    }

    /**
     * Sets the nickname of the contact.
     *
     * @param nickname Nickname of contact.
     */
    public void setNickname(String nickname) {
        Boolean changed = false;
        if (nickname != null) {
            if (this.nickname == null || !this.nickname.equals(nickname)) {
                changed = true;
            }
            this.nickname = nickname;
        }
        else {
            if (this.nickname == null || !this.nickname.equals(getName())) {
                changed = true;
            }
            this.nickname = getName();
        }
        if (changed && getManager().isActivated()) {
            Log.debug("TransportBuddy: Triggering contact update for "+this);
            getManager().getSession().updateContact(this);
        }
    }

    /**
     * Retrieves the groups of the contact.
     *
     * @return Groups contact is in.
     */
    public Collection<String> getGroups() {
       return groups;
    }

    /**
     * Sets the list of groups of the contact.
     *
     * @param groups List of groups the contact is in.
     */
    public void setGroups(List<String> groups) {
        Boolean changed = false;
        if (groups != null && !groups.isEmpty()) {
            if (this.groups == null || this.groups.isEmpty() || !groups.containsAll(this.groups) || !this.groups.containsAll(groups)) {
                changed = true;
            }
            this.groups = groups;
        }
        else {
            if (this.groups != null && !this.groups.isEmpty()) {
                changed = true;
            }
            this.groups = null;
        }
        if (changed && getManager().isActivated()) {
            Log.debug("TransportBuddy: Triggering contact update for "+this);
            getManager().getSession().updateContact(this);
        }
    }

    /**
     * Sets the nickname and list of groups of the contact.
     *
     * @param nickname Nickname of contact.
     * @param groups List of groups the contact is in.
     */
    public void setNicknameAndGroups(String nickname, List<String> groups) {
        Boolean changed = false;
        if (nickname != null) {
            if (this.nickname == null || !this.nickname.equals(nickname)) {
                changed = true;
            }
            this.nickname = nickname;
        }
        else {
            if (this.nickname == null || !this.nickname.equals(getName())) {
                changed = true;
            }
            this.nickname = getName();
        }
        if (groups != null && !groups.isEmpty()) {
            if (this.groups == null || this.groups.isEmpty() || !groups.containsAll(this.groups) || !this.groups.containsAll(groups)) {
                changed = true;
            }
            this.groups = groups;
        }
        else {
            if (this.groups != null && !this.groups.isEmpty()) {
                changed = true;
            }
            this.groups = null;
        }
        if (changed && getManager().isActivated()) {
            Log.debug("TransportBuddy: Triggering contact update for "+this);
            getManager().getSession().updateContact(this);
        }
    }

    /**
     * Retrieves the subscription status for the contact.
     *
     * @return SubType if set.
     */
    public RosterItem.SubType getSubType() {
        return subtype;
    }

    /**
     * Sets the subscription status for the contact.
     *
     * @param substatus Subscription status to be set.
     */
    public void setSubType(RosterItem.SubType substatus) {
        subtype = substatus;
    }

    /**
     * Retrieves the ask status for the contact.
     *
     * @return AskType if set.
     */
    public RosterItem.AskType getAskType() {
        return asktype;
    }

    /**
     * Sets the ask status for the contact.
     *
     * @param askstatus Ask status to be set.
     */
    public void setAskType(RosterItem.AskType askstatus) {
        asktype = askstatus;
    }

    /**
     * Retrieves the current status.
     *
     * @return Current status setting.
     */
    public PresenceType getPresence() {
        return presence;
    }

    /**
     * Sets the current status.
     *
     * @param newpresence New presence to set to.
     */
    public void setPresence(PresenceType newpresence) {
        if (newpresence == null) {
            newpresence = PresenceType.unknown;
        }
        if (newpresence.equals(PresenceType.unavailable)) {
            verboseStatus = "";
        }
        if (!presence.equals(newpresence)) {
            Presence p = new Presence();
            p.setTo(getManager().getSession().getJID());
            p.setFrom(jid);
            getManager().getSession().getTransport().setUpPresencePacket(p, newpresence);
            if (!verboseStatus.equals("")) {
                p.setStatus(verboseStatus);
            }
            getManager().sendPacket(p);
        }
        presence = newpresence;
    }

    /**
     * Retrieves the current verbose status.
     *
     * @return Current verbose status.
     */
    public String getVerboseStatus() {
        return verboseStatus;
    }

    /**
     * Sets the current verbose status.
     *
     * @param newstatus New verbose status.
     */
    public void setVerboseStatus(String newstatus) {
        if (newstatus == null) {
            newstatus = "";
        }
        if (!verboseStatus.equals(newstatus)) {
            Presence p = new Presence();
            p.setTo(getManager().getSession().getJID());
            p.setFrom(jid);
            getManager().getSession().getTransport().setUpPresencePacket(p, presence);
            if (!newstatus.equals("")) {
                p.setStatus(newstatus);
            }
            getManager().sendPacket(p);
        }
        verboseStatus = newstatus;
    }

    /**
     * Convenience routine to set both presence and verbose status at the same time.
     *
     * @param newpresence New presence to set to.
     * @param newstatus New verbose status.
     */
    public void setPresenceAndStatus(PresenceType newpresence, String newstatus) {
        Log.debug("Updating status ["+newpresence+","+newstatus+"] for "+this);
        if (newpresence == null) {
            newpresence = PresenceType.unknown;
        }
        if (newstatus == null) {
            newstatus = "";
        }
        if (newpresence.equals(PresenceType.unavailable)) {
            newstatus = "";
        }
        if (!presence.equals(newpresence) || !verboseStatus.equals(newstatus)) {
            Presence p = new Presence();
            p.setTo(getManager().getSession().getJID());
            p.setFrom(jid);
            getManager().getSession().getTransport().setUpPresencePacket(p, newpresence);
            if (!newstatus.equals("")) {
                p.setStatus(newstatus);
            }
            getManager().sendPacket(p);
        }
        presence = newpresence;
        verboseStatus = newstatus;
    }

    /**
     * Sends the current presence to the session user.
     *
     * @param to JID to send presence updates to.
     */
    public void sendPresence(JID to) {
        Presence p = new Presence();
        p.setTo(to);
        p.setFrom(jid);
        getManager().getSession().getTransport().setUpPresencePacket(p, presence);
        if (verboseStatus != null && verboseStatus.length() > 0) {
            p.setStatus(verboseStatus);
        }
        getManager().sendPacket(p);
    }

    /**
     * Sends the current presence only if it's not unavailable.
     *
     * @param to JID to send presence updates to.
     */
    public void sendPresenceIfAvailable(JID to) {
       if (!presence.equals(PresenceType.unavailable)) {
           sendPresence(to);
       }
    }

    /**
     * Sends an offline presence for a contact only if it is currently online.
     *
     * In case you are wondering why, this is useful when a resource goes offline and we want to indicate all contacts are offline.
     *
     * @param to JID to send presence updates to.
     */
    public void sendOfflinePresenceIfAvailable(JID to) {
       if (!presence.equals(PresenceType.unavailable)) {
           Presence p = new Presence();
           p.setType(Presence.Type.unavailable);
           p.setTo(to);
           p.setFrom(jid);
           getManager().sendPacket(p);
       }
    }

    /**
     * Outputs information about the transport buddy in a pretty format.
     */
    public String toString() {
        return "{Buddy: "+this.jid+" (Nickname: "+this.nickname+") (Groups: "+this.groups+")}";
    }

}
