/**
 * $Revision$
 * $Date$
 *
 * Copyright (C) 2006-2007 Jive Software. All rights reserved.
 *
 * This software is published under the terms of the GNU Public License (GPL),
 * a copy of which is included in this distribution.
 */

package org.jivesoftware.openfire.gateway.web;

import org.jivesoftware.util.PropertyEventListener;
import org.jivesoftware.util.PropertyEventDispatcher;
import org.jivesoftware.util.JiveGlobals;
import org.jivesoftware.openfire.gateway.registration.RegistrationManager;
import org.jivesoftware.openfire.gateway.registration.Registration;
import org.jivesoftware.openfire.gateway.type.TransportType;
import org.jivesoftware.openfire.gateway.GatewayPlugin;
import org.jivesoftware.openfire.XMPPServer;
import org.jivesoftware.openfire.container.PluginManager;
import org.xmpp.packet.JID;
import org.apache.log4j.Logger;

import java.util.*;

/**
 * This provides the conduit through which all XMLRPC interface actions occur.
 *
 * @author Daniel Henninger
 */
public class XMLRPCConduit implements PropertyEventListener {

    static Logger Log = Logger.getLogger(XMLRPCConduit.class);

    ConfigManager configManager = new ConfigManager();

    private String authPassword = null;

    /**
     * Manages registration information.
     * @see org.jivesoftware.openfire.gateway.registration.RegistrationManager
     */
    public final RegistrationManager registrationManager = new RegistrationManager();

    public XMLRPCConduit() {
        authPassword = JiveGlobals.getProperty("plugin.gateway.xmlrpc.password", null);
        PropertyEventDispatcher.addListener(this);
    }

    public void destroy() {
        PropertyEventDispatcher.removeListener(this);
        authPassword = null;
    }

    private boolean verifyPassword(String password) {
        if (authPassword == null) {
            Log.error("XML-RPC authentication not set up on server.");
            return false;
        }
        if (!authPassword.equals(password)) {
            Log.warn("Unauthorized attempt on XML-RPC interface.");
            return false;
        }
        return true;
    }

    /**
     * Toggles whether a transport is enabled or disabled.
     *
     * @param password Auth password for making changes
     * @param transportName Name of the transport to be enabled or disabled (type of transport)
     * @return True or false if the transport is enabled after this call.
     */
    public boolean toggleTransport(String password, String transportName) {
        if (verifyPassword(password)) {
            return configManager.toggleTransport(transportName);
        }
        else {
            return false;
        }
    }

    /**
     * Adds a new registration
     *
     * @param password Auth password for making changes
     * @param user Username or full JID of user who is getting an account registered.
     * @param transportType Type of transport to add user to.
     * @param legacyUsername User's username on the legacy service.
     * @param legacyPassword User's password on the legacy service.
     * @param legacyNickname User's nickname on the legacy service.
     * @return Error message or "Success" on success.
     */
    public String addRegistration(String password, String user, String transportType, String legacyUsername, String legacyPassword, String legacyNickname) {
        if (!verifyPassword(password)) {
            return "Authorization failed!";
        }
        String results = configManager.addRegistration(user, transportType, legacyUsername, legacyPassword, legacyNickname);
        if (results == null) {
            return "Success";
        }
        else {
            return results;
        }
    }

    /**
     * Deletes a registration.
     *
     * @param password Auth password for making changes
     * @param user Username or full JID of user who is getting an account deleted.
     * @param transportType Type of transport to add user to.
     * @return Error message or "Success" on success.
     */
    public String deleteRegistration(String password, String user, String transportType) {
        if (!verifyPassword(password)) {
            return "Authorization failed!";
        }
        JID jid;
        if (user.contains("@")) {
            jid = new JID(user);
        }
        else {
            jid = new JID(user, XMPPServer.getInstance().getServerInfo().getName(), null);
        }
        Collection<Registration> registrations = registrationManager.getRegistrations(jid, TransportType.valueOf(transportType));
        if (registrations.isEmpty()) {
            // User is not registered with us.
            return "Unable to find registration to delete.";
        }
        Registration registration = registrations.iterator().next();
        String results = configManager.deleteRegistration((int)registration.getRegistrationID());
        if (results == null) {
            return "Success";
        }
        else {
            return results;
        }
    }

    /**
     * Updates a registration.
     *
     * @param password Auth password for making changes
     * @param user Username or full JID of user who is getting an account updated.
     * @param transportType Type of transport to add user to.
     * @param legacyUsername User's username on the legacy service.
     * @param legacyPassword User's password on the legacy service.
     * @param legacyNickname User's nickname on the legacy service.
     * @return Error message or "Success" on success.
     */
    public String updateRegistration(String password, String user, String transportType, String legacyUsername, String legacyPassword, String legacyNickname) {
        if (!verifyPassword(password)) {
            return "Authorization failed!";
        }
        JID jid;
        if (user.contains("@")) {
            jid = new JID(user);
        }
        else {
            jid = new JID(user, XMPPServer.getInstance().getServerInfo().getName(), null);
        }
        Collection<Registration> registrations = registrationManager.getRegistrations(jid, TransportType.valueOf(transportType));
        if (registrations.isEmpty()) {
            // User is not registered with us.
            return "Unable to find registration to update.";
        }
        Registration registration = registrations.iterator().next();
        String results = configManager.updateRegistration((int)registration.getRegistrationID(), legacyUsername, legacyPassword, legacyNickname);
        if (results == null) {
            return "Success";
        }
        else {
            return results;
        }
    }

    /**
     * Retrieve a list of all active/enabled transports.
     *
     * @param password Auth password for making changes
     * @return List of active transports.
     */
    public List<String> getActiveTransports(String password) {
        if (!verifyPassword(password)) {
            return Arrays.asList("Authorization failed!");
        }
        PluginManager pluginManager = XMPPServer.getInstance().getPluginManager();
        GatewayPlugin plugin = (GatewayPlugin)pluginManager.getPlugin("gateway");
        List<String> activeTransports = new ArrayList<String>();
        for (String transport : plugin.getTransports()) {
            if (plugin.serviceEnabled(transport)) {
                activeTransports.add(transport);
            }
        }
        return activeTransports;
    }

    public void propertySet(String property, Map<String, Object> params) {
        if (property.equals("plugin.gateway.xmlrpc.password")) {
            authPassword = (String)params.get("value");
        }
    }

    public void propertyDeleted(String property, Map<String, Object> params) {
        if (property.equals("plugin.gateway.xmlrpc.password")) {
            authPassword = null;
        }
    }

    public void xmlPropertySet(String property, Map<String, Object> params) {
        propertySet(property, params);
    }

    public void xmlPropertyDeleted(String property, Map<String, Object> params) {
        propertyDeleted(property, params);
    }
}
